import logging
import sys


def log(message: str, level=logging.INFO):
    logging.log(level, message)


def exit_error(message: str, exit_code: int = 1):
    log(message, level=logging.ERROR)
    raise sys.exit(exit_code)



import fcntl
import os
from contextlib import contextmanager


class FileBasedLock:

    def __init__(self, lock_file_path):
        self._lock_file_path = lock_file_path
        self._lock_fd = None

    def try_lock(self):
        if self._lock_fd is not None:
            raise Exception('Cannot lock twice!')

        fd = os.open(
            self._lock_file_path,
            os.O_RDWR | os.O_CREAT | os.O_APPEND
        )

        try:
            fcntl.flock(
                fd,
                # obtain exclusive lock, fail if unable to lock
                fcntl.LOCK_EX | fcntl.LOCK_NB
            )
        except (IOError, OSError):
            # failed to lock
            os.close(fd)
            return False

        # let's open file again to erase it and store our PID for reference
        with open(self._lock_file_path, 'w') as f:
            f.write(f'PID: {os.getpid()}')

        # lock acquired
        self._lock_fd = fd

        return True

    def release_lock(self):
        self._assert_locked()

        # clear lock file
        open(self._lock_file_path, 'w').close()

        # release lock
        fd = self._lock_fd
        self._lock_fd = None
        fcntl.flock(fd, fcntl.F_UNLCK)
        os.close(fd)

    def _assert_locked(self):
        if self._lock_fd is None:
            raise Exception('Illegal state: not locked')

    @contextmanager
    def acquire_lock(self):
        locked = self.try_lock()
        try:
            yield locked
        finally:
            if locked:
                self.release_lock()


@contextmanager
def acquire_file_lock(lock_file_path):
    lock = FileBasedLock(lock_file_path)
    locked = lock.try_lock()
    try:
        yield locked
    finally:
        if locked:
            lock.release_lock()



import io
import os
from threading import Thread
from typing import Callable

from pyshrimp._internal.pipes.utils import _collect_stream_to_string
from pyshrimp.execution_pipeline.pipeline_api import PipelineElement, PipelineExecutionResult


class _AsyncFunctionPipelineElement(PipelineElement):

    def __init__(self, function: Callable, left_out):
        super().__init__()

        self._left_out = left_out
        self._function = function
        self._result = None
        self._stdout_collected = None
        self._exception = None

        # create output pipe for function
        pipe_r, pipe_w = os.pipe()

        self._open_write_pipe = io.open(pipe_w, 'wb')

        # TODO: add support for binary streams
        self._right_out_writer = io.TextIOWrapper(
            self._open_write_pipe,
            write_through=True,
            line_buffering=False,
            encoding='UTF-8'
        )

        open_read_pipe = io.open(pipe_r, 'rb')
        self._right_out = io.TextIOWrapper(
            open_read_pipe,
            encoding='UTF-8'
        )

        self._thread = Thread(target=self._thread_main)
        self._thread.start()

    def _thread_main(self):
        try:
            self._result = self._function(
                stream_input=self._left_out,
                stream_output=self._right_out_writer
            )
        except Exception as ex:
            self._exception = ex
        finally:
            self._right_out_writer.close()

    def _close_once(self):
        self._thread.join()
        self._stdout_collected = _collect_stream_to_string(self._right_out)
        # TODO: ?? if should_raise and self._exception: raise AsyncFunctionInvocationException(self._exception)

    def stdout_for_pipe(self):
        return self._right_out

    @property
    def result(self) -> PipelineExecutionResult:
        return PipelineExecutionResult(
            result=self._result,
            exception=self._exception,
            stdout=self._stdout_collected,
            stderr=None
        )



# noinspection PyProtectedMember
from pyshrimp._internal.wrapper.mainwrapper import _run as run
from pyshrimp.execution_pipeline.pipeline import pipe, ExecutionPipeline
from pyshrimp.execution_pipeline.pipeline_api import PipelineExecutionResult
from pyshrimp.execution_pipeline.pipeline_starter import PIPE, PIPE_END, PIPE_END_STDOUT
from pyshrimp.utils.command import cmd, shell_cmd, Command, SkipConfig, CommandArgProcessor, DefaultCommandArgProcessor
from pyshrimp.utils.dotdict import as_dot_dict
from pyshrimp.utils.filesystem import ls, glob_ls
from pyshrimp.utils.locking import acquire_file_lock, FileBasedLock
from pyshrimp.utils.logging import log, exit_error
from pyshrimp.utils.matching import re_match_all
from pyshrimp.utils.parallel import in_background
from pyshrimp.utils.string_wrapper import StringWrapper
from pyshrimp.utils.subprocess_utils import run_process, ProcessExecutionException, ProcessExecutionResult
from pyshrimp.utils.wait import wait_until, wait_until_gen



import os
from glob import glob

from pyshrimp.exception import IllegalArgumentException


def _ls(list_supplier, dirs_only=False, files_only=False):
    if dirs_only and files_only:
        raise IllegalArgumentException('Illegal arguments - dirs_only and files_only cannot be true together')

    full_paths = [
        # remove trailing separator, but don't do that if the whole string is trailing separator
        # just in case someone actually uses /** which should return the /
        path.rstrip(os.path.sep) if path != os.path.sep else path for path in list_supplier()
    ]

    if dirs_only or files_only:
        return list(
            filter(
                os.path.isdir if dirs_only else os.path.isfile,
                full_paths
            )
        )
    else:
        # no filtering required
        return full_paths


def ls(*path_segments: str, dirs_only=False, files_only=False):
    path = os.path.join(*path_segments)
    return _ls(
        lambda: [
            os.path.join(path, file_name) for file_name in os.listdir(path)
        ],
        dirs_only=dirs_only,
        files_only=files_only
    )


def glob_ls(*path_glob_segments: str, dirs_only=False, files_only=False, recursive=False):
    return _ls(
        lambda: glob(os.path.join(*path_glob_segments), recursive=recursive),
        dirs_only=dirs_only,
        files_only=files_only
    )



import re
from typing import Union, List


class StringWrapper(str):

    def lines(self, include_empty=False):
        all_lines = self.splitlines()
        if include_empty:
            return all_lines
        else:
            return [line for line in all_lines if line]

    def match_lines(self, pattern, capture_group: Union[str, int] = 1, include_empty_lines=False) -> List[Union[str, None]]:
        match_list = (
            re.match(pattern, el) for el in self.lines(include_empty_lines)
        )
        return [
            m.group(capture_group) for m in match_list if m
        ]

    def match_lines_multi_group(self, pattern, capture_groups: List[Union[str, int]], include_empty_lines=False) -> List[List[Union[str, None]]]:
        match_list = (
            re.match(pattern, el) for el in self.lines(include_empty_lines)
        )
        return [
            [
                m.group(g) for g in capture_groups
            ] for m in match_list if m
        ]

    def columns(self, *column_index, split_pattern=r'\s+', strip_before_split=True, include_empty_lines=False):
        def _process_line(line):
            if strip_before_split:
                line = line.strip()

            split_line = re.split(split_pattern, line)
            return [(split_line[i:i + 1] or [None])[0] for i in column_index]

        return [
            _process_line(line) for line in self.lines(include_empty_lines)
        ]



from dataclasses import dataclass
from typing import Any, List, Tuple

_dot_dict_state_field_name = '_DotDict__dot_dict_state_d40de1e455ae'


def _format_path(path):
    res = []

    for idx, (el_name, sep) in enumerate(path):
        if idx > 0:
            res.append(sep)

        res.append(el_name)

    return ''.join(res)


@dataclass
class _DotDictState:
    data: Any
    path: List[Tuple[str, str]]


def _map_item(dot_dict, p, v):
    if isinstance(v, dict):
        return DotDict(
            data=v,
            path=getattr(dot_dict, _dot_dict_state_field_name).path + p
        )

    elif isinstance(v, list):
        return [_map_item(dot_dict, p + [(f'[{idx}]', '')], el) for idx, el in enumerate(v)]

    else:
        return v


class DotDict:
    """
    Dict wrapper class designed to enable property-like access to dictionary elements.

    Example:

            >>> d = DotDict(original_dict)
            >>> print(d.a.b.c)

    """

    def __init__(self, data, path=None):
        # use some weird name that have very low chance of colliding with actual key in target dict
        # the full version of this name must match the _dot_dict_state_field_name
        self.__dot_dict_state_d40de1e455ae = _DotDictState(
            data=data,
            path=path or []
        )

    def __getattr__(self, item):
        try:
            return _map_item(self, [(item, '.')], self.__dot_dict_state_d40de1e455ae.data[item])
        except KeyError:
            raise KeyError(_format_path(self.__dot_dict_state_d40de1e455ae.path + [(item, '.')]))

    def __setattr__(self, key, value):
        if key == _dot_dict_state_field_name:
            super().__setattr__(key, value)
        else:
            self.__dot_dict_state_d40de1e455ae.data[key] = value

    def __repr__(self):
        return repr(self.__dot_dict_state_d40de1e455ae.data)

    def __str__(self):
        return str(self.__dot_dict_state_d40de1e455ae.data)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)


def as_dot_dict(dict_data, variable_name='dict'):
    return DotDict(dict_data, [(variable_name, '.')])



import logging
import time
from typing import Callable, TypeVar

T = TypeVar('T')
_log = logging.getLogger()


def wait_until_gen(
    waiting_until: str,
    timeout_seconds=60,
    sleep_interval_sec=1
):
    _log.info(f'Waiting until {waiting_until} (max wait: {timeout_seconds}s)')
    timeout_at = time.time() + timeout_seconds - sleep_interval_sec

    try:
        step = 0
        while True:
            yield step

            if time.time() >= timeout_at:
                _log.warning(f' ! WARNING: got timeout while waiting until {waiting_until}')
                return False

            time.sleep(sleep_interval_sec)
            step += 1

    except GeneratorExit:
        _log.info(f' + OK')


def wait_until(
    waiting_until: str,
    collect: Callable[[], T],
    timeout_sec=60,
    check_interval_sec=1,
    before_check=lambda res: None,
    before_sleep=lambda res: None,
    check=lambda res: bool(res),
    on_timeout=lambda res: None,
    silent=False
) -> (bool, T):
    if not silent:
        _log.info(f'Waiting until {waiting_until} (max wait: {timeout_sec}s)')

    timeout_at = time.time() + timeout_sec - check_interval_sec

    while True:
        res = collect()

        before_check(res)

        if check(res):
            if not silent:
                _log.info(f' + OK')
                _log.info(f'')

            return True, res
        else:
            if time.time() >= timeout_at:
                if not silent:
                    _log.warning(f' ! WARNING: got timeout while waiting until {waiting_until}')
                    _log.info(f'')

                on_timeout(res)
                return False, res

        before_sleep(res)
        time.sleep(check_interval_sec)



# noinspection PyMethodMayBeStatic
from dataclasses import dataclass
from typing import Optional, Any

# noinspection PyProtectedMember
from pyshrimp._internal.pipes.utils import _collect_stream_to_string


@dataclass
class PipelineExecutionResult:
    stdout: Optional[str]
    stderr: Optional[str]
    result: Optional[Any]
    exception: Optional[Exception]


class PipelineElement:

    def __init__(self):
        self._closed = False

    def _close_once(self):
        pass

    def close(self):
        if not self._closed:
            self._close_once()
            self._closed = True

    def stdout_for_pipe(self):
        raise NotImplementedError()

    @property
    def result(self) -> PipelineExecutionResult:
        raise NotImplementedError


class StringPipelineElement(PipelineElement):

    def __init__(self, out: str) -> None:
        super().__init__()
        self._out = out

    def stdout_for_pipe(self):
        return self._out

    @property
    def result(self) -> PipelineExecutionResult:
        return PipelineExecutionResult(
            result=None,
            stdout=self._out,
            stderr=None,
            exception=None
        )


class StreamPipelineElement(PipelineElement):

    def __init__(self, out) -> None:
        super().__init__()
        self._out = out
        self._out_collected = None

    def stdout_for_pipe(self):
        return self._out

    def _close_once(self):
        self._out_collected = _collect_stream_to_string(self._out)

    @property
    def result(self) -> PipelineExecutionResult:
        return PipelineExecutionResult(
            result=None,
            stdout=self._out_collected,
            stderr=None,
            exception=None
        )


class PipelineTerminator:
    pass


class PipelineTerminatorStdout:
    pass
